@@ -38,6 +38,16 @@ push(char *bfr, int size, MMIOT *f)
 }
 
 
+/*
+ * push a character into the generator input buffer
+ */
+static void 
+pushc(char c, MMIOT *f)
+{
+    EXPAND(f->in) = c;
+}
+
+
 /* look <i> characters ahead of the cursor.
  */
 static inline int
@@ -46,16 +56,16 @@ peek(MMIOT *f, int i)
 
     i += (f->isp-1);
 
-    return (i >= 0) && (i < S(f->in)) ? T(f->in)[i] : EOF;
+    return (i >= 0) && (i < S(f->in)) ? (unsigned char)T(f->in)[i] : EOF;
 }
 
 
 /* pull a byte from the input buffer
  */
-static inline int
+static inline unsigned int
 pull(MMIOT *f)
 {
-    return ( f->isp < S(f->in) ) ? T(f->in)[f->isp++] : EOF;
+    return ( f->isp < S(f->in) ) ? (unsigned char)T(f->in)[f->isp++] : EOF;
 }
 
 
@@ -73,7 +83,11 @@ isthisspace(MMIOT *f, int i)
 {
     int c = peek(f, i);
 
-    return isspace(c) || (c == EOF);
+    if ( c == EOF )
+	return 1;
+    if ( c & 0x80 )
+	return 0;
+    return isspace(c) || (c < ' ');
 }
 
 
@@ -94,8 +108,10 @@ isthisnonword(MMIOT *f, int i)
 
 
 /* return/set the current cursor position
+ * (when setting the current cursor position we also need to flush the
+ * last character written cache)
  */
-#define mmiotseek(f,x)	(f->isp = x)
+#define mmiotseek(f,x)	((f->isp = x), (f->last = 0))
 #define mmiottell(f)	(f->isp)
 
 
@@ -164,6 +180,16 @@ Qprintf(MMIOT *f, char *fmt, ...)
 }
 
 
+/* Qanchor() prints out a suitable-for-id-tag version of a string
+ */
+static void
+Qanchor(struct line *p, MMIOT *f)
+{
+    mkd_string_to_anchor(T(p->text), S(p->text),
+			 (mkd_sta_function_t)Qchar, f, 1, f);
+}
+
+
 /* Qem()
  */
 static void
@@ -183,9 +209,10 @@ Qem(MMIOT *f, char c, int count)
 /* generate html from a markup fragment
  */
 void
-___mkd_reparse(char *bfr, int size, int flags, MMIOT *f)
+___mkd_reparse(char *bfr, int size, mkd_flag_t flags, MMIOT *f, char *esc)
 {
     MMIOT sub;
+    struct escaped e;
 
     ___mkd_initmmiot(&sub, f->footnotes);
     
@@ -193,19 +220,49 @@ ___mkd_reparse(char *bfr, int size, int flags, MMIOT *f)
     sub.cb = f->cb;
     sub.ref_prefix = f->ref_prefix;
 
+    if ( esc ) {
+	sub.esc = &e;
+	e.up = f->esc;
+	e.text = esc;
+    }
+    else
+	sub.esc = f->esc;
+
     push(bfr, size, &sub);
-    EXPAND(sub.in) = 0;
+    pushc(0, &sub);
     S(sub.in)--;
     
     text(&sub);
     ___mkd_emblock(&sub);
     
     Qwrite(T(sub.out), S(sub.out), f);
+    /* inherit the last character printed from the reparsed
+     * text;  this way superscripts can work when they're
+     * applied to something embedded in a link
+     */
+    f->last = sub.last;
 
     ___mkd_freemmiot(&sub, f->footnotes);
 }
 
 
+/*
+ * check the escape list for special cases
+ */
+static int
+escaped(MMIOT *f, char c)
+{
+    struct escaped *thing = f->esc;
+
+    while ( thing ) {
+	if ( strchr(thing->text, c) )
+	    return 1;
+	thing = thing->up;
+    }
+    return 0;
+}
+
+
 /*
  * write out a url, escaping problematic characters
  */
@@ -232,7 +289,7 @@ puturl(char *s, int size, MMIOT *f, int display)
 	    Qstring("%22", f);
 	else if ( isalnum(c) || ispunct(c) || (display && isspace(c)) )
 	    Qchar(c, f);
-	else if ( c == 003 )	/* untokenize ^C */
+	else if ( c == MKD_EOLN )	/* untokenize hard return */
 	    Qstring("  ", f);
 	else
 	    Qprintf(f, "%%%02X", c);
@@ -416,7 +473,7 @@ linkyurl(MMIOT *f, int image, Footnote *p)
 
     if ( c == '<' ) {
 	pull(f);
-	if ( !(f->flags & MKD_1_COMPAT) )
+	if ( !is_flag_set(f->flags, MKD_1_COMPAT) )
 	    return linkybroket(f,image,p);
 	mayneedtotrim=1;
     }
@@ -539,7 +596,7 @@ printlinkyref(MMIOT *f, linkytype *tag, char *link, int size)
 {
     char *edit;
     
-    if ( f->flags & IS_LABEL )
+    if ( is_flag_set(f->flags, IS_LABEL) )
 	return;
     
     Qstring(tag->link_pfx, f);
@@ -553,7 +610,7 @@ printlinkyref(MMIOT *f, linkytype *tag, char *link, int size)
 	    puturl(link + tag->szpat, size - tag->szpat, f, 0);
     }
     else
-	___mkd_reparse(link + tag->szpat, size - tag->szpat, MKD_TAGTEXT, f);
+	___mkd_reparse(link + tag->szpat, size - tag->szpat, MKD_TAGTEXT, f, 0);
 
     Qstring(tag->link_sfx, f);
 
@@ -585,10 +642,10 @@ extra_linky(MMIOT *f, Cstring text, Footnote *ref)
 	return 0;
 	
     if ( f->flags & IS_LABEL )
-    	___mkd_reparse(T(text), S(text), linkt.flags, f);
+    	___mkd_reparse(T(text), S(text), linkt.flags, f, 0);
     else {
 	ref->flags |= REFERENCED;
-	ref->refnumber = ++ f->reference;
+	ref->refnumber = ++ f->footnotes->reference;
 	Qprintf(f, "<sup id=\"%sref:%d\"><a href=\"#%s:%d\" rel=\"footnote\">%d</a></sup>",
 		p_or_nothing(f), ref->refnumber,
 		p_or_nothing(f), ref->refnumber, ref->refnumber);
@@ -597,6 +654,32 @@ extra_linky(MMIOT *f, Cstring text, Footnote *ref)
 } /* extra_linky */
 
 
+
+/* check a url (or url fragment to see that it begins with a known good
+ * protocol (or no protocol at all)
+ */
+static int
+safelink(Cstring link)
+{
+    char *p, *colon;
+
+    if ( T(link) == 0 )	/* no link; safe */
+	return 1;
+
+    p = T(link);
+    if ( (colon = memchr(p, ':', S(link))) == 0 )
+	return 1; /* no protocol specified: safe */
+
+    if ( !isalpha(*p) )	/* protocol/method is [alpha][alnum or '+.-'] */
+	return 1;
+    while ( ++p < colon )
+	if ( !(isalnum(*p) || *p == '.' || *p == '+' || *p == '-') )
+	    return 1;
+
+    return isautoprefix(T(link), S(link));
+}
+
+
 /* print out a linky (or fail if it's Not Allowed)
  */
 static int
@@ -604,15 +687,14 @@ linkyformat(MMIOT *f, Cstring text, int image, Footnote *ref)
 {
     linkytype *tag;
 
-    if ( image || (ref == 0) )
+
+    if ( image )
 	tag = &imaget;
     else if ( tag = pseudo(ref->link) ) {
-	if ( f->flags & (MKD_NO_EXT|MKD_SAFELINK) )
+	if ( is_flag_set(f->flags, MKD_NO_EXT) || is_flag_set(f->flags, MKD_SAFELINK) )
 	    return 0;
     }
-    else if ( (f->flags & MKD_SAFELINK) && T(ref->link)
-				        && (T(ref->link)[0] != '/')
-				        && !isautoprefix(T(ref->link), S(ref->link)) )
+    else if ( is_flag_set(f->flags, MKD_SAFELINK) && !safelink(ref->link) )
 	/* if MKD_SAFELINK, only accept links that are local or
 	 * a well-known protocol
 	 */
@@ -623,8 +705,8 @@ linkyformat(MMIOT *f, Cstring text, int image, Footnote *ref)
     if ( f->flags & tag->flags )
 	return 0;
 
-    if ( f->flags & IS_LABEL )
-	___mkd_reparse(T(text), S(text), tag->flags, f);
+    if ( is_flag_set(f->flags, IS_LABEL) )
+	___mkd_reparse(T(text), S(text), tag->flags, f, 0);
     else if ( tag->link_pfx ) {
 	printlinkyref(f, tag, T(ref->link), S(ref->link));
 
@@ -635,12 +717,12 @@ linkyformat(MMIOT *f, Cstring text, int image, Footnote *ref)
 
 	if ( S(ref->title) ) {
 	    Qstring(" title=\"", f);
-	    ___mkd_reparse(T(ref->title), S(ref->title), MKD_TAGTEXT, f);
+	    ___mkd_reparse(T(ref->title), S(ref->title), MKD_TAGTEXT, f, 0);
 	    Qchar('"', f);
 	}
 
 	Qstring(tag->text_pfx, f);
-	___mkd_reparse(T(text), S(text), tag->flags, f);
+	___mkd_reparse(T(text), S(text), tag->flags, f, 0);
 	Qstring(tag->text_sfx, f);
     }
     else
@@ -687,9 +769,9 @@ linkylinky(int image, MMIOT *f)
 		 * require a second []
 		 */
 		mmiotseek(f, implicit_mark);
-		goodlink = !(f->flags & MKD_1_COMPAT);
+		goodlink = !is_flag_set(f->flags, MKD_1_COMPAT);
 
-		if ( (f->flags & MKD_EXTRA_FOOTNOTE) && (!image) && S(name) && T(name)[0] == '^' )
+		if ( is_flag_set(f->flags, MKD_EXTRA_FOOTNOTE) && (!image) && S(name) && T(name)[0] == '^' )
 		    extra_footnote = 1;
 	    }
 	    
@@ -700,15 +782,14 @@ linkylinky(int image, MMIOT *f)
 		    S(key.tag) = S(name);
 		}
 
-		if ( ref = bsearch(&key, T(*f->footnotes), S(*f->footnotes),
-					  sizeof key, (stfu)__mkd_footsort) ) {
+		if ( ref = bsearch(&key, T(f->footnotes->note),
+					 S(f->footnotes->note),
+					 sizeof key, (stfu)__mkd_footsort) ) {
 		    if ( extra_footnote )
 			status = extra_linky(f,name,ref);
 		    else
 			status = linkyformat(f, name, image, ref);
 		}
-		else if ( f->flags & IS_LABEL )
-		    status = linkyformat(f, name, image, 0);
 	    }
 	}
     }
@@ -745,8 +826,12 @@ static void
 mangle(char *s, int len, MMIOT *f)
 {
     while ( len-- > 0 ) {
+#if DEBIAN_GLITCH
+	Qprintf(f, "&#%02d;", *((unsigned char*)(s++)) );
+#else
 	Qstring("&#", f);
 	Qprintf(f, COINTOSS() ? "x%02x;" : "%02d;", *((unsigned char*)(s++)) );
+#endif
     }
 }
 
@@ -808,20 +893,21 @@ code(MMIOT *f, char *s, int length)
     int i,c;
 
     for ( i=0; i < length; i++ )
-	if ( (c = s[i]) == 003)  /* ^C: expand back to 2 spaces */
+	if ( (c = s[i]) == MKD_EOLN)  /* expand back to 2 spaces */
 	    Qstring("  ", f);
+	else if ( c == '\\' && (i < length-1) && escaped(f, s[i+1]) )
+	    cputc(s[++i], f);
 	else
 	    cputc(c, f);
 } /* code */
 
-
 /*  delspan() -- write out a chunk of text, blocking with <del>...</del>
  */
 static void
 delspan(MMIOT *f, int size)
 {
     Qstring("<del>", f);
-    ___mkd_reparse(cursor(f)-1, size, 0, f);
+    ___mkd_reparse(cursor(f)-1, size, 0, f, 0);
     Qstring("</del>", f);
 }
 
@@ -851,12 +937,12 @@ forbidden_tag(MMIOT *f)
 {
     int c = toupper(peek(f, 1));
 
-    if ( f->flags & MKD_NOHTML )
+    if ( is_flag_set(f->flags, MKD_NOHTML) )
 	return 1;
 
-    if ( c == 'A' && (f->flags & MKD_NOLINKS) && !isthisalnum(f,2) )
+    if ( c == 'A' && is_flag_set(f->flags, MKD_NOLINKS) && !isthisalnum(f,2) )
 	return 1;
-    if ( c == 'I' && (f->flags & MKD_NOIMAGE)
+    if ( c == 'I' && is_flag_set(f->flags, MKD_NOIMAGE)
 		  && strncasecmp(cursor(f)+1, "MG", 2) == 0
 		  && !isthisalnum(f,4) )
 	return 1;
@@ -903,7 +989,7 @@ process_possible_link(MMIOT *f, int size)
     int mailto = 0;
     char *text = cursor(f);
     
-    if ( f->flags & MKD_NOLINKS ) return 0;
+    if ( is_flag_set(f->flags, MKD_NOLINKS) ) return 0;
 
     if ( (size > 7) && strncasecmp(text, "mailto:", 7) == 0 ) {
 	/* if it says it's a mailto, it's a mailto -- who am
@@ -950,7 +1036,7 @@ maybe_tag_or_link(MMIOT *f)
     int c, size;
     int maybetag = 1;
 
-    if ( f->flags & MKD_TAGTEXT )
+    if ( is_flag_set(f->flags, MKD_TAGTEXT) )
 	return 0;
 
     for ( size=0; (c = peek(f, size+1)) != '>'; size++) {
@@ -963,11 +1049,9 @@ maybe_tag_or_link(MMIOT *f)
 	}
 	else if ( isspace(c) )
 	    break;
-#if WITH_GITHUB_TAGS
-	else if ( ! (c == '/' || c == '-' || c == '_' || isalnum(c) ) )
-#else
-	else if ( ! (c == '/' || isalnum(c) ) )
-#endif
+	else if ( ! (c == '/'
+		     || (is_flag_set(f->flags, MKD_GITHUBTAGS) && (c == '-' || c == '_'))
+		     || isalnum(c) ) )
 	    maybetag=0;
     }
 
@@ -1013,13 +1097,16 @@ maybe_autolink(MMIOT *f)
 
     /* greedily scan forward for the end of a legitimate link.
      */
-    for ( size=0; (c=peek(f, size+1)) != EOF; size++ )
+    for ( size=0; (c=peek(f, size+1)) != EOF; size++ ) {
 	if ( c == '\\' ) {
 	     if ( peek(f, size+2) != EOF )
 		++size;
 	}
-	else if ( isspace(c) || strchr("'\"()[]{}<>`", c) )
+	else if ( c & 0x80 )	/* HACK: ignore utf-8 extended characters */
+	    continue;
+	else if ( isspace(c) || strchr("'\"()[]{}<>`", c) || c == MKD_EOLN )
 	    break;
+    }
 
     if ( (size > 1) && process_possible_link(f, size) ) {
 	shift(f, size);
@@ -1058,7 +1145,7 @@ islike(MMIOT *f, char *s)
     int len;
     int i;
 
-    if ( s[0] == '<' ) {
+    if ( s[0] == '|' ) {
 	if ( !isthisnonword(f, -1) )
 	    return 0;
        ++s;
@@ -1067,7 +1154,7 @@ islike(MMIOT *f, char *s)
     if ( !(len = strlen(s)) )
 	return 0;
 
-    if ( s[len-1] == '>' ) {
+    if ( s[len-1] == '|' ) {
 	if ( !isthisnonword(f,len-1) )
 	    return 0;
 	len--;
@@ -1086,13 +1173,13 @@ static struct smarties {
     char *entity;
     int shift;
 } smarties[] = {
-    { '\'', "'s>",      "rsquo",  0 },
-    { '\'', "'t>",      "rsquo",  0 },
-    { '\'', "'re>",     "rsquo",  0 },
-    { '\'', "'ll>",     "rsquo",  0 },
-    { '\'', "'ve>",     "rsquo",  0 },
-    { '\'', "'m>",      "rsquo",  0 },
-    { '\'', "'d>",      "rsquo",  0 },
+    { '\'', "'s|",      "rsquo",  0 },
+    { '\'', "'t|",      "rsquo",  0 },
+    { '\'', "'re|",     "rsquo",  0 },
+    { '\'', "'ll|",     "rsquo",  0 },
+    { '\'', "'ve|",     "rsquo",  0 },
+    { '\'', "'m|",      "rsquo",  0 },
+    { '\'', "'d|",      "rsquo",  0 },
     { '-',  "---",      "mdash",  2 },
     { '-',  "--",       "ndash",  1 },
     { '.',  "...",      "hellip", 2 },
@@ -1100,11 +1187,11 @@ static struct smarties {
     { '(',  "(c)",      "copy",   2 },
     { '(',  "(r)",      "reg",    2 },
     { '(',  "(tm)",     "trade",  3 },
-    { '3',  "<3/4>",    "frac34", 2 },
-    { '3',  "<3/4ths>", "frac34", 2 },
-    { '1',  "<1/2>",    "frac12", 2 },
-    { '1',  "<1/4>",    "frac14", 2 },
-    { '1',  "<1/4th>",  "frac14", 2 },
+    { '3',  "|3/4|",    "frac34", 2 },
+    { '3',  "|3/4ths|", "frac34", 2 },
+    { '1',  "|1/2|",    "frac12", 2 },
+    { '1',  "|1/4|",    "frac14", 2 },
+    { '1',  "|1/4th|",  "frac14", 2 },
     { '&',  "&#0;",      0,       3 },
 } ;
 #define NRSMART ( sizeof smarties / sizeof smarties[0] )
@@ -1117,7 +1204,9 @@ smartypants(int c, int *flags, MMIOT *f)
 {
     int i;
 
-    if ( f->flags & (MKD_NOPANTS|MKD_TAGTEXT|IS_LABEL) )
+    if ( is_flag_set(f->flags, MKD_NOPANTS) 
+      || is_flag_set(f->flags, MKD_TAGTEXT)
+      || is_flag_set(f->flags, IS_LABEL) )
 	return 0;
 
     for ( i=0; i < NRSMART; i++)
@@ -1146,7 +1235,7 @@ smartypants(int c, int *flags, MMIOT *f)
 			    break;
 			else if ( c == '\'' && peek(f, j+1) == '\'' ) {
 			    Qstring("&ldquo;", f);
-			    ___mkd_reparse(cursor(f)+1, j-2, 0, f);
+			    ___mkd_reparse(cursor(f)+1, j-2, 0, f, 0);
 			    Qstring("&rdquo;", f);
 			    shift(f,j+1);
 			    return 1;
@@ -1161,6 +1250,27 @@ smartypants(int c, int *flags, MMIOT *f)
 } /* smartypants */
 
 
+/* process latex with arbitrary 2-character ( $$ .. $$, \[ .. \], \( .. \)
+ * delimiters
+ */
+static int
+mathhandler(MMIOT *f, int e1, int e2)
+{
+    int i = 0;
+
+    while(peek(f, ++i) != EOF) {
+        if (peek(f, i) == e1 && peek(f, i+1) == e2) {
+	    cputc(peek(f,-1), f);
+	    cputc(peek(f, 0), f);
+	    while ( i-- > -1 )
+		cputc(pull(f), f);
+            return 1;
+        }
+    }
+    return 0;
+}
+
+
 /* process a body of text encased in some sort of tick marks.   If it
  * works, generate the output and return 1, otherwise just return 0 and
  * let the caller figure it out.
@@ -1188,7 +1298,7 @@ tickhandler(MMIOT *f, int tickchar, int minticks, int allow_space, spanhandler s
     return 0;
 }
 
-#define tag_text(f)	(f->flags & MKD_TAGTEXT)
+#define tag_text(f)	is_flag_set(f->flags, MKD_TAGTEXT)
 
 
 static void
@@ -1199,7 +1309,7 @@ text(MMIOT *f)
     int smartyflags = 0;
 
     while (1) {
-        if ( (f->flags & MKD_AUTOLINK) && isalpha(peek(f,1)) && !tag_text(f) )
+        if ( is_flag_set(f->flags, MKD_AUTOLINK) && isalpha(peek(f,1)) && !tag_text(f) )
 	    maybe_autolink(f);
 
         c = pull(f);
@@ -1212,7 +1322,8 @@ text(MMIOT *f)
 	switch (c) {
 	case 0:     break;
 
-	case 3:     Qstring(tag_text(f) ? "  " : "<br/>", f);
+	case MKD_EOLN:
+		    Qstring(tag_text(f) ? "  " : "<br/>", f);
 		    break;
 
 	case '>':   if ( tag_text(f) )
@@ -1235,13 +1346,18 @@ text(MMIOT *f)
 		    else
 			Qchar(c, f);
 		    break;
+
 	case '[':   if ( tag_text(f) || !linkylinky(0, f) )
 			Qchar(c, f);
 		    break;
 	/* A^B -> A<sup>B</sup> */
-	case '^':   if ( (f->flags & (MKD_NOSUPERSCRIPT|MKD_STRICT|MKD_TAGTEXT))
-				|| (isthisnonword(f,-1) && peek(f,-1) != ')')
-				|| isthisspace(f,1) )
+	case '^':   if ( is_flag_set(f->flags, MKD_NOSUPERSCRIPT)
+			    || is_flag_set(f->flags, MKD_STRICT)
+			    || is_flag_set(f->flags, MKD_TAGTEXT)
+			    || (f->last == 0)
+			    || ((ispunct(f->last) || isspace(f->last))
+						    && f->last != ')')
+			    || isthisspace(f,1) )
 			Qchar(c,f);
 		    else {
 			char *sup = cursor(f);
@@ -1268,15 +1384,14 @@ text(MMIOT *f)
 			    shift(f,len);
 			}
 			Qstring("<sup>",f);
-			___mkd_reparse(sup, len, 0, f);
+			___mkd_reparse(sup, len, 0, f, "()");
 			Qstring("</sup>", f);
 		    }
 		    break;
 	case '_':
 	/* Underscores don't count if they're in the middle of a word */
-		    if ( !(f->flags & (MKD_NORELAXED|MKD_STRICT))
-					&& isthisalnum(f,-1)
-					 && isthisalnum(f,1) ) {
+		    if ( !(is_flag_set(f->flags, MKD_NORELAXED) || is_flag_set(f->flags, MKD_STRICT))
+				&& isthisalnum(f,-1) && isthisalnum(f,1) ) {
 			Qchar(c, f);
 			break;
 		    }
@@ -1297,7 +1412,10 @@ text(MMIOT *f)
 		    }
 		    break;
 	
-	case '~':   if ( (f->flags & (MKD_NOSTRIKETHROUGH|MKD_TAGTEXT|MKD_STRICT)) || ! tickhandler(f,c,2,0, delspan) )
+	case '~':   if ( is_flag_set(f->flags, MKD_NOSTRIKETHROUGH)
+			 || is_flag_set(f->flags, MKD_STRICT)
+			 || is_flag_set(f->flags, MKD_TAGTEXT)
+			 || ! tickhandler(f,c,2,0, delspan) )
 			Qchar(c, f);
 		    break;
 
@@ -1308,9 +1426,19 @@ text(MMIOT *f)
 	case '\\':  switch ( c = pull(f) ) {
 		    case '&':   Qstring("&amp;", f);
 				break;
-		    case '<':   Qstring("&lt;", f);
+		    case '<':   c = peek(f,1);
+				if ( (c == EOF) || isspace(c) )
+				    Qstring("&lt;", f);
+				else {
+				    /* Markdown.pl does not escape <[nonwhite]
+				     * sequences */
+				    Qchar('\\', f);
+				    shift(f, -1);
+				}
+				
 				break;
-		    case '^':   if ( f->flags & (MKD_STRICT|MKD_NOSUPERSCRIPT) ) {
+		    case '^':   if ( is_flag_set(f->flags, MKD_STRICT)
+					|| is_flag_set(f->flags, MKD_NOSUPERSCRIPT) ) {
 				    Qchar('\\', f);
 				    shift(f,-1);
 				    break;
@@ -1319,7 +1447,7 @@ text(MMIOT *f)
 				break;
 				
 		    case ':': case '|':
-				if ( f->flags & MKD_NOTABLES ) {
+				if ( is_flag_set(f->flags, MKD_NOTABLES) ) {
 				    Qchar('\\', f);
 				    shift(f,-1);
 				    break;
@@ -1327,16 +1455,22 @@ text(MMIOT *f)
 				Qchar(c, f);
 				break;
 				
-		    case '>': case '#': case '.': case '-':
-		    case '+': case '{': case '}': case ']':
-		    case '!': case '[': case '*': case '_':
-		    case '\\':case '(': case ')':
-		    case '`':	Qchar(c, f);
+		    case EOF:	Qchar('\\', f);
 				break;
-		    default:
-				Qchar('\\', f);
-				if ( c != EOF )
-				    shift(f,-1);
+
+		    case '[':
+		    case '(':   if ( is_flag_set(f->flags, MKD_LATEX)
+				   && mathhandler(f, '\\', (c =='(')?')':']') )
+				    break;
+				/* else fall through to default */
+			
+		    default:    if ( escaped(f,c) ||
+				     strchr(">#.-+{}]![*_\\()`", c) )
+				    Qchar(c, f);
+				else {
+				    Qchar('\\', f);
+				    shift(f, -1);
+				}
 				break;
 		    }
 		    break;
@@ -1355,7 +1489,16 @@ text(MMIOT *f)
 			Qchar(c, f);
 		    break;
 
-	default:    Qchar(c, f);
+	case '$':   if ( is_flag_set(f->flags, MKD_LATEX) && (peek(f, 1) == '$') ) {
+			pull(f);
+			if ( mathhandler(f, '$', '$') )
+			    break;
+			Qchar('$', f);
+		    }
+		    /* fall through to default */
+	
+	default:    f->last = c;
+		    Qchar(c, f);
 		    break;
 	}
     }
@@ -1369,26 +1512,22 @@ text(MMIOT *f)
 static void
 printheader(Paragraph *pp, MMIOT *f)
 {
-#if WITH_ID_ANCHOR
-    Qprintf(f, "<h%d", pp->hnumber);
-    if ( f->flags & MKD_TOC ) {
-	Qstring(" id=\"", f);
-	mkd_string_to_anchor(T(pp->text->text),
-			     S(pp->text->text),
-			     (mkd_sta_function_t)Qchar, f, 1);
-	Qchar('"', f);
+    if ( is_flag_set(f->flags, MKD_IDANCHOR) ) {
+	Qprintf(f, "<h%d", pp->hnumber);
+	if ( is_flag_set(f->flags, MKD_TOC) ) {
+	    Qstring(" id=\"", f);
+	    Qanchor(pp->text, f);
+	    Qchar('"', f);
+	}
+	Qchar('>', f);
+    } else {
+	if ( is_flag_set(f->flags, MKD_TOC) ) {
+	    Qstring("<a name=\"", f);
+	    Qanchor(pp->text, f);
+	    Qstring("\"></a>\n", f);
+	}
+	Qprintf(f, "<h%d>", pp->hnumber);
     }
-    Qchar('>', f);
-#else
-    if ( f->flags & MKD_TOC ) {
-	Qstring("<a name=\"", f);
-	mkd_string_to_anchor(T(pp->text->text),
-			     S(pp->text->text),
-			     (mkd_sta_function_t)Qchar, f, 1);
-	Qstring("\"></a>\n", f);
-    }
-    Qprintf(f, "<h%d>", pp->hnumber);
-#endif
     push(T(pp->text->text), S(pp->text->text), f);
     text(f);
     Qprintf(f, "</h%d>", pp->hnumber);
@@ -1397,20 +1536,20 @@ printheader(Paragraph *pp, MMIOT *f)
 
 enum e_alignments { a_NONE, a_CENTER, a_LEFT, a_RIGHT };
 
-static char* alignments[] = { "", " align=\"center\"", " align=\"left\"",
-				  " align=\"right\"" };
+static char* alignments[] = { "", " style=\"text-align:center;\"",
+				  " style=\"text-align:left;\"",
+				  " style=\"text-align:right;\"" };
 
 typedef STRING(int) Istring;
 
 static int
-splat(Line *p, int leading_pipe, char *block, Istring align, int force, MMIOT *f)
+splat(Line *p, char *block, Istring align, int force, MMIOT *f)
 {
     int first,
 	idx = p->dle,
 	colno = 0;
 
 
-    if ( leading_pipe ) idx++;
     ___mkd_tidy(&p->text);
     if ( T(p->text)[S(p->text)-1] == '|' )
 	--S(p->text);
@@ -1430,7 +1569,7 @@ splat(Line *p, int leading_pipe, char *block, Istring align, int force, MMIOT *f
 	Qprintf(f, "<%s%s>",
 		   block,
 		   alignments[ (colno < S(align)) ? T(align)[colno] : a_NONE ]);
-	___mkd_reparse(T(p->text)+first, idx-first, 0, f);
+	___mkd_reparse(T(p->text)+first, idx-first, 0, f, "|");
 	Qprintf(f, "</%s>\n", block);
 	idx++;
 	colno++;
@@ -1452,20 +1591,27 @@ printtable(Paragraph *pp, MMIOT *f)
 
     Line *hdr, *dash, *body;
     Istring align;
-    int hcols,start,starts_with_pipe=0;
+    int hcols,start;
     char *p;
+    enum e_alignments it;
 
     hdr = pp->text;
     dash= hdr->next;
     body= dash->next;
 
-    starts_with_pipe = T(hdr->text)[hdr->dle] == '|';
+    if ( T(hdr->text)[hdr->dle] == '|' ) {
+	/* trim leading pipe off all lines
+	 */
+	Line *r;
+	for ( r = pp->text; r; r = r->next )
+	    r->dle ++;
+    }
 
-    /* first figure out cell alignments */
+    /* figure out cell alignments */
 
     CREATE(align);
 
-    for (p=T(dash->text), start=0; start < S(dash->text); ) {
+    for (p=T(dash->text), start=dash->dle; start < S(dash->text); ) {
 	char first, last;
 	int end;
 	
@@ -1478,14 +1624,16 @@ printtable(Paragraph *pp, MMIOT *f)
 		last = p[end];
 	    }
 	}
-	EXPAND(align) = ( first == ':' ) ? (( last == ':') ? a_CENTER : a_LEFT)
-					 : (( last == ':') ? a_RIGHT : a_NONE );
+	it = ( first == ':' ) ? (( last == ':') ? a_CENTER : a_LEFT)
+			      : (( last == ':') ? a_RIGHT : a_NONE );
+
+	EXPAND(align) = it;
 	start = 1+end;
     }
 
     Qstring("<table>\n", f);
     Qstring("<thead>\n", f);
-    hcols = splat(hdr, starts_with_pipe, "th", align, 0, f);
+    hcols = splat(hdr, "th", align, 0, f);
     Qstring("</thead>\n", f);
 
     if ( hcols < S(align) )
@@ -1496,7 +1644,7 @@ printtable(Paragraph *pp, MMIOT *f)
 
     Qstring("<tbody>\n", f);
     for ( ; body; body = body->next)
-	splat(body, starts_with_pipe, "td", align, 1, f);
+	splat(body, "td", align, 1, f);
     Qstring("</tbody>\n", f);
     Qstring("</table>\n", f);
 
@@ -1508,9 +1656,10 @@ printtable(Paragraph *pp, MMIOT *f)
 static int
 printblock(Paragraph *pp, MMIOT *f)
 {
-    Line *t = pp->text;
     static char *Begin[] = { "", "<p>", "<p style=\"text-align:center;\">"  };
     static char *End[]   = { "", "</p>","</p>" };
+    Line *t = pp->text;
+    int align = pp->align;
 
     while (t) {
 	if ( S(t->text) ) {
@@ -1518,30 +1667,73 @@ printblock(Paragraph *pp, MMIOT *f)
 			 && T(t->text)[S(t->text)-2] == ' '
 			 && T(t->text)[S(t->text)-1] == ' ' ) {
 		push(T(t->text), S(t->text)-2, f);
-		push("\003\n", 2, f);
+		pushc(MKD_EOLN, f);
+		pushc('\n', f);
 	    }
 	    else {
 		___mkd_tidy(&t->text);
 		push(T(t->text), S(t->text), f);
 		if ( t->next )
-		    push("\n", 1, f);
+		    pushc('\n', f);
 	    }
 	}
 	t = t->next;
     }
-    Qstring(Begin[pp->align], f);
+    Qstring(Begin[align], f);
     text(f);
-    Qstring(End[pp->align], f);
+    Qstring(End[align], f);
     return 1;
 }
 
 
 static void
-printcode(Line *t, MMIOT *f)
+printcode(Line *t, char *lang, MMIOT *f)
 {
     int blanks;
 
-    Qstring("<pre><code>", f);
+    if ( f->cb->e_codefmt ) {
+	/* external code block formatter;  copy the text into a buffer,
+	 * call the formatter to style it, then dump that styled text
+	 * directly to the queue
+	 */
+	char *text;
+	char *fmt;
+	int size, copy_p;
+	Line *p;
+
+	for (size=0, p = t; p; p = p->next )
+	    size += 1+S(p->text);
+
+	text = malloc(1+size);
+
+	for ( copy_p = 0; t ; t = t->next ) {
+	    memcpy(text+copy_p, T(t->text), S(t->text));
+	    copy_p += S(t->text);
+	    text[copy_p++] = '\n';
+	}
+	text[copy_p] = 0;
+	
+	fmt = (*(f->cb->e_codefmt))(text, copy_p, (lang && lang[0]) ? lang : 0);
+	free(text);
+
+	if ( fmt ) {
+	    Qwrite(fmt, strlen(fmt), f);
+	    if ( f->cb->e_free )
+		(*(f->cb->e_free))(fmt, f->cb->e_data);
+	    return;
+	}
+	/* otherwise the external formatter failed and we need to 
+	 * fall back to the traditional codeblock format
+	 */
+    }
+    
+    Qstring("<pre><code", f);
+    if (lang && lang[0]) {
+      Qstring(" class=\"", f);
+      Qstring(lang, f);
+      Qstring("\"", f);
+    }
+    Qstring(">", f);
     for ( blanks = 0; t ; t = t->next ) {
 	if ( S(t->text) > t->dle ) {
 	    while ( blanks ) {
@@ -1576,17 +1768,57 @@ printhtml(Line *t, MMIOT *f)
 
 
 static void
-htmlify(Paragraph *p, char *block, char *arguments, MMIOT *f)
+htmlify_paragraphs(Paragraph *p, MMIOT *f)
 {
     ___mkd_emblock(f);
-    if ( block )
-	Qprintf(f, arguments ? "<%s %s>" : "<%s>", block, arguments);
-    ___mkd_emblock(f);
 
     while (( p = display(p, f) )) {
 	___mkd_emblock(f);
 	Qstring("\n\n", f);
     }
+}
+
+
+#ifdef GITHUB_CHECKBOX
+static void
+li_htmlify(Paragraph *p, char *arguments, mkd_flag_t flags, MMIOT *f)
+{
+    ___mkd_emblock(f);
+
+    Qprintf(f, "<li");
+    if ( arguments )
+	Qprintf(f, " %s", arguments);
+    if ( flags & GITHUB_CHECK )
+	Qprintf(f, " class=\"github_checkbox\"");
+    Qprintf(f, ">");
+#if CHECKBOX_AS_INPUT
+    if ( flags & GITHUB_CHECK ) {
+	Qprintf(f, "<input disabled=\"\" type=\"checkbox\"");
+	if ( flags & IS_CHECKED )
+	    Qprintf(f, " checked=\"checked\"");
+	Qprintf(f, "/>");
+    }
+#else
+    if ( flags & GITHUB_CHECK )
+	Qprintf(f, flags & IS_CHECKED ? "&#x2611;" : "&#x2610;");
+#endif
+
+    htmlify_paragraphs(p, f);
+
+     Qprintf(f, "</li>");
+    ___mkd_emblock(f);
+}
+#endif
+
+
+static void
+htmlify(Paragraph *p, char *block, char *arguments, MMIOT *f)
+{
+    ___mkd_emblock(f);
+    if ( block )
+	Qprintf(f, arguments ? "<%s %s>" : "<%s>", block, arguments);
+
+    htmlify_paragraphs(p, f);
 
     if ( block )
 	 Qprintf(f, "</%s>", block);
@@ -1605,7 +1837,7 @@ definitionlist(Paragraph *p, MMIOT *f)
 	for ( ; p ; p = p->next) {
 	    for ( tag = p->text; tag; tag = tag->next ) {
 		Qstring("<dt>", f);
-		___mkd_reparse(T(tag->text), S(tag->text), 0, f);
+		___mkd_reparse(T(tag->text), S(tag->text), 0, f, 0);
 		Qstring("</dt>\n", f);
 	    }
 
@@ -1628,7 +1860,11 @@ listdisplay(int typ, Paragraph *p, MMIOT* f)
 	Qprintf(f, ">\n");
 
 	for ( ; p ; p = p->next ) {
+#ifdef GITHUB_CHECKBOX
+	    li_htmlify(p->down, p->ident, p->flags, f);
+#else
 	    htmlify(p->down, "li", p->ident, f);
+#endif
 	    Qchar('\n', f);
 	}
 
@@ -1654,7 +1890,7 @@ display(Paragraph *p, MMIOT *f)
 	break;
 	
     case CODE:
-	printcode(p->text, f);
+	printcode(p->text, p->lang, f);
 	break;
 	
     case QUOTE:
@@ -1703,21 +1939,21 @@ mkd_extra_footnotes(MMIOT *m)
     int j, i;
     Footnote *t;
 
-    if ( m->reference == 0 )
+    if ( m->footnotes->reference == 0 )
 	return;
 
     Csprintf(&m->out, "\n<div class=\"footnotes\">\n<hr/>\n<ol>\n");
     
-    for ( i=1; i <= m->reference; i++ ) {
-	for ( j=0; j < S(*m->footnotes); j++ ) {
-	    t = &T(*m->footnotes)[j];
+    for ( i=1; i <= m->footnotes->reference; i++ ) {
+	for ( j=0; j < S(m->footnotes->note); j++ ) {
+	    t = &T(m->footnotes->note)[j];
 	    if ( (t->refnumber == i) && (t->flags & REFERENCED) ) {
-		Csprintf(&m->out, "<li id=\"%s:%d\">\n<p>",
+		Csprintf(&m->out, "<li id=\"%s:%d\">\n",
 			    p_or_nothing(m), t->refnumber);
-		Csreparse(&m->out, T(t->title), S(t->title), 0);
+		htmlify(t->text, 0, 0, m);
 		Csprintf(&m->out, "<a href=\"#%sref:%d\" rev=\"footnote\">&#8617;</a>",
 			    p_or_nothing(m), t->refnumber);
-		Csprintf(&m->out, "</p></li>\n");
+		Csprintf(&m->out, "</li>\n");
 	    }
 	}
     }
@@ -1736,19 +1972,22 @@ mkd_document(Document *p, char **res)
     if ( p && p->compiled ) {
 	if ( ! p->html ) {
 	    htmlify(p->code, 0, 0, p->ctx);
-	    if ( p->ctx->flags & MKD_EXTRA_FOOTNOTE )
+	    if ( is_flag_set(p->ctx->flags, MKD_EXTRA_FOOTNOTE) )
 		mkd_extra_footnotes(p->ctx);
 	    p->html = 1;
-	}
-
-	size = S(p->ctx->out);
+	    size = S(p->ctx->out);
 	
-	if ( (size == 0) || T(p->ctx->out)[size-1] )
-	    EXPAND(p->ctx->out) = 0;
+	    if ( (size == 0) || T(p->ctx->out)[size-1] ) {
+		/* Add a null byte at the end of the generated html,
+		 * but pretend it doesn't exist.
+		 */
+		EXPAND(p->ctx->out) = 0;
+		--S(p->ctx->out);
+	    }
+	}
 	
 	*res = T(p->ctx->out);
-	return size;
+	return S(p->ctx->out);
     }
     return EOF;
 }
-
